/* Copyright © 2010 Brandon L Black <blblack@gmail.com> and Jay Reitz <jreitz@gmail.com>
 *
 * This file is part of gdnsd-plugin-georeg.
 *
 * gdnsd-plugin-georeg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * gdnsd-plugin-georeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with gdnsd.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "config.h"
#define GDNSD_PLUGIN_NAME georeg
#include <gdnsd-plugin.h>

#include <GeoIP.h>
#include <GeoIPCity.h>
#include <string.h>
#include <math.h>

typedef struct {
    double lat;
    double lon;
} coord_t;

typedef struct {
    const char* name;
    coord_t coord;
} dc_t;

typedef struct {
    unsigned dcnum;
    double distance;
} dist_sort_t;

typedef struct {
    uint32_t addr;
    monio_state_t state;
} addrstate_t;

typedef struct {
    addrstate_t* as;
    unsigned count;
} addrstates_t;

typedef union {
    uint8_t* cname;
    addrstates_t a;
} resource_dc_t;

typedef struct {
    const char* name;
    // A list of all datacenter numbers in preference order
    //  for defaulting when no location data is available.
    uint8_t* def_order;

    // mappings is indexed by datacenter number, but only
    //  contains valid config data in the members corresponding
    //  to valid[x] == true.
    bool* valid;
    resource_dc_t* mappings;

    // was this resource initialized with addr or cname data?
    bool is_cname;
} resource_t;

static unsigned num_dcs = 0;
static unsigned num_res = 0;
static dc_t* dcs = NULL;
static resource_t* resources = NULL;
static GeoIP* geoip = NULL;
static bool global_failover = true;

monio_list_t monio_list = { 0, NULL };

// Geographic distance between two lat/long points.
// Because we only care about rough distance comparison rather than
//  the precise values themselves, input is specified in radians
//  and output in units of the earth's diameter. DEG2RAD here is
//  the conversion factor, since our config data is in degrees.
static const double DEG2RAD = 0.017453292519943295769236907684886;
F_CONST
static double haversine(double lat1, double lon1, double lat2, double lon2) {
    double a = pow(sin((lat2 - lat1) * 0.5), 2.0)
        + cos(lat1) * cos(lat2) * pow(sin((lon2 - lon1) * 0.5), 2.0);
    return atan2(sqrt(a), sqrt(1.0 - a));
}

// Yes, I know about double equality and epsilon, but
//  for these purposes, it simply doesn't matter.
F_PURE F_NONNULL
static int dist_sort(const void* a, const void* b) {
    assert(a); assert(b);
    const dist_sort_t* da = (const dist_sort_t*) a;
    const dist_sort_t* db = (const dist_sort_t*) b;
    if(da->distance < db->distance)
        return -1;
    else if(da->distance > db->distance)
        return 1;
    else
        return 0;
}

// The formula is (x[0] - 48) + ((x[1] - 48) << 6),
//  which results in mapping two characters of the range 0-Z (48 to 90)
//  down to two numbers from 0-42, and then concatenating them as 6-bit
//  numbers to produce a maximum range of 0 - 2730 (2730 is the encoding of "ZZ").
// The final index is multiplied by the number of datacenters to find
//  an offset in dcs_by_whatever below.
// NUM_CODES is used to size these arrays.
#define NUM_CODES 2731
static uint8_t* dcs_by_country = NULL;
static uint8_t* dcs_by_us_state = NULL;
static uint8_t* dcs_by_ca_prov = NULL;

F_PURE F_WUNUSED F_NONNULL
static unsigned int code_to_offset(const char* code) {
    assert(code);
    const uint8_t* ucode = (const uint8_t*)code;
    if( ucode[0] > 47 && ucode[0] < 91
     && ucode[1] > 47 && ucode[1] < 91)
        return ((ucode[0] - 48) + ((ucode[1] - 48) << 6)) * num_dcs;
    else
        log_fatal("plugin_georeg: Illegal location (sub)code '%s'", ucode);
}

F_PURE F_NONNULL
static uint8_t* code_to_local_dclist(const char* code, unsigned code_len) {
    if(code_len == 2)
        return &dcs_by_country[code_to_offset(code)];
    else if(code_len == 4 && code[0] == 'U' && code[1] == 'S')
        return &dcs_by_us_state[code_to_offset(&code[2])];
    else if(code_len == 4 && code[0] == 'C' && code [1] == 'A')
        return &dcs_by_ca_prov[code_to_offset(&code[2])];
    else
        log_fatal("plugin_georeg: Illegal location code '%s'", code);
}

#define CFG_OPT_BOOL(_data, _name, _store) \
    do { \
        const vscf_data_t* _setting = vscf_hash_get_data_byconstkey(_data, _name, true); \
        if(_setting) { \
            if(!vscf_is_simple(_setting) \
            || !vscf_simple_get_as_bool(_setting, &_store)) \
                log_fatal("plugin_georeg: option %s: Value must be 'true' or 'false'", _name); \
        } \
    } while(0)

static bool bad_key(const char* key, unsigned klen, const vscf_data_t* d V_UNUSED, void* data) {
    log_fatal(data, key, klen);
}

static void plugin_georeg_monio_add(const char* svctype, const char* resname, const char* addr_txt, monio_state_t* state_ptr) {
    monio_list.info = realloc(monio_list.info, sizeof(monio_info_t) * (monio_list.count + 1));
    monio_info_t* m = &monio_list.info[monio_list.count++];
    
    m->svctype = svctype;
    m->desc = resname;
    m->addr = addr_txt;
    m->state_ptr = state_ptr;
}

monio_list_t* plugin_georeg_load_config(const vscf_data_t* config) {
    assert(vscf_get_type(config) == VSCF_HASH_T);

    // Open the GeoIP database
    const vscf_data_t* geoip_fn = vscf_hash_get_data_byconstkey(config, "geoip_database", true);
    if(!geoip_fn) {
        geoip_fn = vscf_hash_get_data_byconstkey(config, "geoip_region_database", true);
        if(geoip_fn)
            log_warn("plugin_georeg: geoip_region_database is deprecated, please use geoip_database instead");
        else
            log_fatal("plugin_georeg: geoip_database must be specified");
    }
    if(VSCF_SIMPLE_T != vscf_get_type(geoip_fn))
        log_fatal("plugin_georeg: geoip_database must be a simple string");
    geoip = GeoIP_open(vscf_simple_get_data(geoip_fn), GEOIP_MEMORY_CACHE);
    if(!geoip) log_fatal("Failed to load GeoIP database");
    if(geoip->databaseType != GEOIP_COUNTRY_EDITION
       && geoip->databaseType != GEOIP_CITY_EDITION_REV1
       && geoip->databaseType != GEOIP_REGION_EDITION_REV1)
        log_fatal("GeoIP database format not supported (Must be Country, Region v1, or City v1)");

    // Get global_failover setting, if any
    CFG_OPT_BOOL(config, "global_failover", global_failover);

    // Set up our complex data...

    const vscf_data_t* loc_coords;
    const vscf_data_t* overrides = NULL;

    const char* loc_fn = DEF_LOCS;
    const vscf_data_t* locations = vscf_hash_get_data_byconstkey(config, "locations", true);
    if(locations) {
        if(VSCF_HASH_T != vscf_get_type(locations))
            log_fatal("plugin_georeg: locations must be a hash");

        const vscf_data_t* loc_file = vscf_hash_get_data_byconstkey(locations, "file", true);
        if(loc_file) {
            if(VSCF_SIMPLE_T != vscf_get_type(loc_file))
                log_fatal("plugin_georeg: locations file must be a simple string");
            loc_fn = vscf_simple_get_data(loc_file);
        }
        overrides = vscf_hash_get_data_byconstkey(locations, "overrides", true);
        if(overrides)
            if(VSCF_HASH_T != vscf_get_type(overrides))
                log_fatal("plugin_georeg: overrides must be a hash");
        
        vscf_hash_iterate(locations, true, bad_key, (void*)"plugin_georeg: Invalid locations key: '%R'");
    }

    const vscf_data_t* datacenters = vscf_hash_get_data_byconstkey(config, "datacenters", true);
    if(!datacenters)
        log_fatal("plugin_georeg: datacenters hash missing");
    if(VSCF_HASH_T != vscf_get_type(datacenters))
        log_fatal("plugin_georeg: datacenters must be a hash");

    num_dcs = vscf_hash_get_len(datacenters);
    if(num_dcs < 2)
        log_fatal("plugin_georeg: you must define 2 or more datacenters");
    if(num_dcs > 255)
        log_fatal("plugin_georeg: too many datacenters defined");

    dcs = calloc(num_dcs, sizeof(dc_t));
    dcs_by_country = calloc(1, NUM_CODES * num_dcs);
    dcs_by_us_state = calloc(1, NUM_CODES * num_dcs);
    dcs_by_ca_prov = calloc(1, NUM_CODES * num_dcs);

    for(unsigned i = 0; i < num_dcs; i++) {
        dc_t* this_dc = &dcs[i];
        this_dc->name = vscf_hash_get_key_byindex(datacenters, i, NULL);
        const vscf_data_t* coord = vscf_hash_get_data_byindex(datacenters, i);

        if( VSCF_ARRAY_T != vscf_get_type(coord)
          || (2 != vscf_array_get_len(coord))
          || VSCF_SIMPLE_T != vscf_get_type(vscf_array_get_data(coord, 0))
          || !vscf_simple_get_as_double(vscf_array_get_data(coord, 0), &this_dc->coord.lon)
          || VSCF_SIMPLE_T != vscf_get_type(vscf_array_get_data(coord, 1))
          || !vscf_simple_get_as_double(vscf_array_get_data(coord, 1), &this_dc->coord.lat))
            log_fatal("plugin_georeg: datacenter '%s' must have an array of 2 floats for coordinates", this_dc->name);

        this_dc->coord.lon *= DEG2RAD;
        this_dc->coord.lat *= DEG2RAD;
    }

    char* vscf_err;
    loc_coords = vscf_scan_filename(loc_fn, &vscf_err);
    if(!loc_coords || VSCF_HASH_T != vscf_get_type(loc_coords))
        log_fatal("Failed to load a hash of location co-ordinates from '%s': %s", loc_fn, vscf_err);

    for(unsigned i = 0; i < vscf_hash_get_len(loc_coords); i++) {
        coord_t loc_coord;
        unsigned code_len;
        const char* code = vscf_hash_get_key_byindex(loc_coords, i, &code_len);
        const vscf_data_t* coord_data = vscf_hash_get_data_byindex(loc_coords, i);
        if( VSCF_ARRAY_T != vscf_get_type(coord_data)
          || (2 != vscf_array_get_len(coord_data))
          || VSCF_SIMPLE_T != vscf_get_type(vscf_array_get_data(coord_data, 0))
          || !vscf_simple_get_as_double(vscf_array_get_data(coord_data, 0), &loc_coord.lon)
          || VSCF_SIMPLE_T != vscf_get_type(vscf_array_get_data(coord_data, 1))
          || !vscf_simple_get_as_double(vscf_array_get_data(coord_data, 1), &loc_coord.lat))
            log_fatal("plugin_georeg: location '%s' must have an array of 2 floats for coordinates", code);

        loc_coord.lon *= DEG2RAD;
        loc_coord.lat *= DEG2RAD;

        uint8_t* local_dclist = code_to_local_dclist(code, code_len);

        dist_sort_t distances[num_dcs];
        for(unsigned i = 0; i < num_dcs; i++) {
            distances[i].dcnum = i;
            distances[i].distance = haversine(loc_coord.lat, loc_coord.lon, dcs[i].coord.lat, dcs[i].coord.lon);
        }
        qsort(distances, num_dcs, sizeof(dist_sort_t), dist_sort);
        for(unsigned i = 0; i < num_dcs; i++)
            local_dclist[i] = distances[i].dcnum;
    }

    vscf_destroy(loc_coords);

    if(overrides) {
        for(unsigned i = 0; i < vscf_hash_get_len(overrides); i++) {
            unsigned code_len;
            const char* code = vscf_hash_get_key_byindex(overrides, i, &code_len);
            const vscf_data_t* loc_override = vscf_hash_get_data_byindex(overrides, i);
            if(VSCF_ARRAY_T != vscf_get_type(loc_override)
              || num_dcs != vscf_array_get_len(loc_override))
                log_fatal("plugin_georeg: override for location '%s' must be an array of all datacenter names in some order", code);

            uint8_t* local_dclist = code_to_local_dclist(code, code_len);

            for(unsigned i = 0; i < num_dcs; i++) {
                const vscf_data_t* dcname_data = vscf_array_get_data(loc_override, i);
                if(VSCF_SIMPLE_T != vscf_get_type(dcname_data))
                    log_fatal("plugin_georeg: overrides for location '%s' must be strings", code);
                const char* dcname = vscf_simple_get_data(dcname_data);
                unsigned dcname_len = vscf_simple_get_len(dcname_data);
                signed index = vscf_hash_get_index_bykey(datacenters, dcname, dcname_len);
                if(index < 0)
                    log_fatal("plugin_georeg: datacenter '%s' invalid in list for location '%s'", dcname, code);
                local_dclist[i] = index;
            }
        }
    }

    const vscf_data_t* res_data = vscf_hash_get_data_byconstkey(config, "resources", true);
    if(!res_data || VSCF_HASH_T != vscf_get_type(res_data))
        log_fatal("plugin_georeg: resources must be specified as a hash of one or more resources");
    num_res = vscf_hash_get_len(res_data);
    if(!num_res)
        log_fatal("plugin_georeg: one or more resources must be defined");

    resources = malloc(num_res * sizeof(resource_t));

    for(unsigned i = 0; i < num_res; i++) {
        resource_t* this_res = &resources[i];
        unsigned res_name_len;
        this_res->name = strdup(vscf_hash_get_key_byindex(res_data, i, &res_name_len));
        const vscf_data_t* res_props = vscf_hash_get_data_byindex(res_data, i);
        if(vscf_get_type(res_props) != VSCF_HASH_T)
            log_fatal("plugin_georeg: the value of resource '%s' must be a hash", this_res->name);

        const char* svctype = NULL;
        const vscf_data_t* res_stype = vscf_hash_get_data_byconstkey(res_props, "service_type", true);
        if(res_stype) {
            if(vscf_get_type(res_stype) != VSCF_SIMPLE_T)
                log_fatal("plugin_georeg: resource '%s': 'service_type' must be a string if defined", this_res->name);
            svctype = vscf_simple_get_data(res_stype);
        }

        const vscf_data_t* addrs_data = vscf_hash_get_data_byconstkey(res_props, "addrs", true);
        const vscf_data_t* cnames_data = vscf_hash_get_data_byconstkey(res_props, "cnames", true);

        unsigned num_mappings = 0;
        if (addrs_data && cnames_data)
            log_fatal("plugin_georeg: resource '%s': resource must be address or cname, not both", this_res->name);
        else if (!addrs_data && !cnames_data)
            log_fatal("plugin_georeg: resource '%s': must contain either 'addrs' or 'cnames' hash of datacenters to resources", this_res->name);
        else if(addrs_data && (VSCF_HASH_T != vscf_get_type(addrs_data) || !(num_mappings = vscf_hash_get_len(addrs_data))))
            log_fatal("plugin_georeg: resource '%s': 'addrs' must be defined as a hash of one or more datacenters to address mappings", this_res->name);
        else if(cnames_data && (VSCF_HASH_T != vscf_get_type(cnames_data) || !(num_mappings = vscf_hash_get_len(cnames_data))))
            log_fatal("plugin_georeg: resource '%s': 'cnames' must be defined as a hash of one or more datacenters to cname mappings", this_res->name);

        if(addrs_data) this_res->is_cname = false;
        else this_res->is_cname = true;
            
        this_res->def_order = calloc(num_dcs, sizeof(uint8_t));
        this_res->valid = calloc(num_dcs, sizeof(bool));
        this_res->mappings = calloc(num_dcs, sizeof(resource_dc_t));

        for(unsigned j = 0; j < num_mappings; j++) {
            unsigned dcname_len;
            const char* dcname = vscf_hash_get_key_byindex(addrs_data ? addrs_data : cnames_data, j, &dcname_len);
            signed dcidx = vscf_hash_get_index_bykey(datacenters, dcname, dcname_len);
            if(dcidx < 0)
                log_fatal("plugin_georeg: resource '%s': invalid datacenter name '%s'", this_res->name, dcname);

            if(addrs_data) {
                const vscf_data_t* addr_data = vscf_hash_get_data_byindex(addrs_data, j);
                assert(addr_data); // byindex should always succeed given hash_get_len-based index
                this_res->def_order[j] = dcidx;
                this_res->valid[dcidx] = true;
                resource_dc_t* this_resdc = &this_res->mappings[dcidx];
                if(VSCF_SIMPLE_T == vscf_get_type(addr_data)) {
                    const char* addr_txt = vscf_simple_get_data(addr_data);
                    struct in_addr a;
                    if(inet_pton(AF_INET, addr_txt, &a) < 1)
                        log_fatal("plugin_georeg: resource '%s': parsing '%s' as IPv4 address failed", this_res->name, addr_txt);
                    this_resdc->a.count = 1;
                    this_resdc->a.as = calloc(1, sizeof(addrstate_t));
                    this_resdc->a.as[0].addr = a.s_addr;

                    // setup monitoring 
                    char * desc = malloc(res_name_len + dcname_len + 2);
                    sprintf(desc, "%s/%s", this_res->name, dcname);
                    plugin_georeg_monio_add(svctype, desc, addr_txt, &this_resdc->a.as[0].state);
                }
                else if(VSCF_HASH_T == vscf_get_type(addr_data)) {
                    unsigned num_lbs = vscf_hash_get_len(addr_data);
                    if(!num_lbs) 
                        log_fatal("plugin_georeg: resource '%s': addrs hash for datacenter '%s' cannot be empty", this_res->name, dcname);
                    this_resdc->a.count = num_lbs;
                    this_resdc->a.as = calloc(num_lbs, sizeof(addrstate_t));
                    for(unsigned k = 0; k < num_lbs; k++) {
                        unsigned lb_name_len;
                        const char* lb_name = vscf_hash_get_key_byindex(addr_data, k, &lb_name_len);
                        const vscf_data_t* lb_data = vscf_hash_get_data_byindex(addr_data, k);
                        if(VSCF_SIMPLE_T != vscf_get_type(lb_data))
                            log_fatal("plugin_georeg: resource '%s': datacenter '%s': name: '%s': The values in the addrs hash must be IPv4 address strings", this_res->name, dcname, lb_name);

                        const char* addr_txt = vscf_simple_get_data(lb_data);
                        struct in_addr a;
                        if(inet_pton(AF_INET, addr_txt, &a) < 1)
                            log_fatal("plugin_georeg: resource '%s': datacenter '%s': name: '%s': parsing '%s' as IPv4 address failed", this_res->name, dcname, lb_name, addr_txt);

                        addrstate_t* this_as = &this_resdc->a.as[k];
                        this_as->addr = a.s_addr;

                        char* desc = malloc(res_name_len + dcname_len + lb_name_len + 3);
                        sprintf(desc, "%s/%s/%s", this_res->name, dcname, lb_name);
                        plugin_georeg_monio_add(svctype, desc, addr_txt, &this_as->state);
                    }
                }
                else {
                    log_fatal("plugin_georeg: resource '%s': value for datacenter '%s' must be either an IPv4 address string or a mapping of logical names to IPv4 addresses", this_res->name, dcname);
                }
            }
            else { // cname data
                const vscf_data_t* cname_data = vscf_hash_get_data_byindex(cnames_data, j);
                if(!cname_data || VSCF_SIMPLE_T != vscf_get_type(cname_data))
                    log_fatal("plugin_georeg: resource '%s': value for datacenter '%s' must be a CNAME string", this_res->name, dcname);
                uint8_t* dname = malloc(256);
                if(vscf_simple_get_as_dname(cname_data, dname) == DNAME_INVALID)
                    log_fatal("plugin_georeg: resource '%s': CNAME for datacenter '%s' is not a legal domainname", this_res->name, dcname);
                dname_trim(dname);

                this_res->mappings[dcidx].cname = dname;
                this_res->def_order[j] = dcidx;
                this_res->valid[dcidx] = true;
            }   
        }

        // If this resource doesn't have addresses in all datacenters, the "valid" array will
        //  take care of most things except for "def_order", which must be expanded with the
        //  remaining unique datacenter indices.
        if(num_mappings < num_dcs) {
            uint8_t* def_order_ptr = &this_res->def_order[num_mappings];
            for(unsigned j = 0; j < num_dcs; j++)
                if(!this_res->valid[j])
                    *def_order_ptr++ = j;
            assert(def_order_ptr == &this_res->def_order[num_dcs]);
        }

        vscf_hash_iterate(res_props, true, bad_key, (void*)"plugin_georeg: Invalid resource option: '%R'");
    }

    vscf_hash_iterate(config, true, bad_key, (void*)"plugin_georeg: Invalid configuration key: '%R'");
    
    return &monio_list;
}

unsigned plugin_georeg_map_resource_dyna(const char* resname) {
    for(unsigned i = 0; i < num_res; i++)
        if(!strcmp(resname, resources[i].name)) {
            if(resources[i].is_cname)
                log_fatal("plugin_georeg: Resource '%s' defined as CNAME data, but used in a DYNA RR", resources[i].name);
            return i;
        }
    log_fatal("plugin_georeg: Invalid resource name '%s' detected from zonefile lookup", resname);
}

unsigned plugin_georeg_map_resource_dync(const char* resname, const uint8_t* origin) {
    for(unsigned i = 0; i < num_res; i++) {
        if(!strcmp(resname, resources[i].name)) {
            if(!resources[i].is_cname)
                log_fatal("plugin_georeg: Resource '%s' defined as address data, but used in a DYNC RR", resources[i].name);
            for(unsigned j = 0; j < num_dcs; j++) {
                if(!resources[i].valid[j])
                    continue;
                const uint8_t* dname = resources[i].mappings[j].cname;
                if(dname_status(dname) == DNAME_PARTIAL) {
                    uint8_t dnbuf[256];
                    dname_copy(dnbuf, dname);
                    if(dname_cat(dnbuf, origin) != DNAME_VALID)
                        log_fatal("plugin_georeg: Name '%s' of resource '%s', when used at origin '%s', produces an invalid domainname", logf_dname(dname), resources[i].name, logf_dname(origin));
                }
            }
            return i;
        }
    }
    log_fatal("plugin_georeg: Invalid resource name '%s' detected from zonefile lookup", resname);
}

static const uint8_t* _anysin_to_dclist(const anysin_t* client, const resource_t* res) {
    const uint8_t* dclist = NULL;


    if(client->sa.sa_family == AF_INET) {

        if(geoip->databaseType == GEOIP_CITY_EDITION_REV1) {
            GeoIPRecord *record;
            record = GeoIP_record_by_ipnum(geoip, ntohl(client->sin.sin_addr.s_addr));
            if(record) {
                assert(record->country_code);
                if(record->country_code[0] && record->country_code[0] != '-') {
                    if(record->region && record->region[0]) {
                        if(record->country_code[0] == 'U' && record->country_code[1] == 'S') {
                            dclist = &dcs_by_us_state[code_to_offset(record->region)];
                        }
                        else if(record->country_code[0] == 'C' && record->country_code[1] == 'A') {
                            dclist = &dcs_by_ca_prov[code_to_offset(record->region)];
                        }
                        else {
                            dclist = &dcs_by_country[code_to_offset(record->country_code)];
                        }
                    }
                    else {
                        dclist = &dcs_by_country[code_to_offset(record->country_code)];
                    }
                }
                GeoIPRecord_delete(record);
            }
        } else if (geoip->databaseType == GEOIP_REGION_EDITION_REV1) {
            GeoIPRegion gipr;
            GeoIP_assign_region_by_inetaddr(geoip, client->sin.sin_addr.s_addr, &gipr);
            if(gipr.region[0]) {
                if(gipr.country_code[0] == 'U' && gipr.country_code[1] == 'S') {
                    dclist = &dcs_by_us_state[code_to_offset(gipr.region)];
                }
                else {
                    assert(gipr.country_code[0] == 'C' && gipr.country_code[1] == 'A');
                    dclist = &dcs_by_ca_prov[code_to_offset(gipr.region)];
                }
            }
            else if(gipr.country_code[0] && gipr.country_code[0] != '-') {
                dclist = &dcs_by_country[code_to_offset(gipr.country_code)];
            }
        }
        else {
            assert(geoip->databaseType == GEOIP_COUNTRY_EDITION);
            const char* country_code = GeoIP_country_name_by_ipnum(geoip, client->sin.sin_addr.s_addr);
            if(country_code)
                dclist = &dcs_by_country[code_to_offset(country_code)];
        }
    }


    // If both of the first two entries are zero, then this is
    //  an un-initialized slot (as in, we have no locations data
    //  for this country/region), so fall back to def_order
    if(!dclist || (!dclist[0] && !dclist[1]))
        dclist = res->def_order;

    return dclist;
}

void plugin_georeg_resolve_dynaddr(unsigned threadnum V_UNUSED, unsigned resnum, const anysin_t* client, dynaddr_result_t* result) {
    const resource_t* r = &resources[resnum];
    assert(!r->is_cname);

    const uint8_t* dclist = _anysin_to_dclist(client, r);

    result->count_v6 = 0;

    // It's simpler to split the code on the global_failover flag
    //  for clarity, even if it's a little redundant...
    if(!global_failover) {
        // Get the index of the first valid datacenter, it's the
        //  only one we'll use.
        uint8_t dcnum = 0;
        for(unsigned i = 0; i < num_dcs; i++) {
            if(r->valid[dclist[i]]) {
                dcnum = dclist[i];
                break;
            }
        }

        const addrstates_t* dcaddrs = &r->mappings[dcnum].a;

        // Try to return only the non-DOWN addresses in the this datacenter
        for(unsigned i = 0; i < dcaddrs->count; i++)
            if(monio_state_get(&dcaddrs->as[i].state) != MONIO_STATE_DOWN)
                result->addrs_v4[result->count_v4++] = dcaddrs->as[i].addr;

        // Failing that, return them all
        if(!result->count_v4)
            for(unsigned i = 0; i < dcaddrs->count; i++)
                result->addrs_v4[result->count_v4++] = dcaddrs->as[i].addr;
     }
     else {
        // Try to return all non-DOWN addresses from the first datacenter
        //  on the preference list which has any non-DOWN addresses.
        for(unsigned i = 0; i < num_dcs; i++) {
            const uint8_t dcnum = dclist[i];
            if(r->valid[dcnum]) {
                const addrstates_t* dcaddrs = &r->mappings[dcnum].a;
                for(unsigned i = 0; i < dcaddrs->count; i++)
                    if(monio_state_get(&dcaddrs->as[i].state) != MONIO_STATE_DOWN)
                        result->addrs_v4[result->count_v4++] = dcaddrs->as[i].addr;
                if(result->count_v4)
                    return;
            }
        }

        // All addresses at all datacenters were down, re-iterate and return
        //  the addresses of the first valid datacenter regardless of state
        for(unsigned i = 0; i < num_dcs; i++) {
            const uint8_t dcnum = dclist[i];
            if(r->valid[dcnum]) {
                const addrstates_t* dcaddrs = &r->mappings[dcnum].a;
                for(unsigned i = 0; i < dcaddrs->count; i++)
                    result->addrs_v4[result->count_v4++] = dcaddrs->as[i].addr;
                return;
            }
        }

        assert(0); // unreachable, because all resources have
                   //  at least one valid datacenter+address
    }
}

void plugin_georeg_resolve_dyncname(unsigned threadnum V_UNUSED, unsigned resnum V_UNUSED, const uint8_t* origin, const anysin_t* client V_UNUSED, dyncname_result_t* result) {
    const resource_t* r = &resources[resnum];
    assert(r->is_cname);

    const uint8_t* dclist = _anysin_to_dclist(client, r);

    // return the first cname (monitoring cnames NYI (is it even reasonable to do so)?)
    for(unsigned i = 0; i < num_dcs; i++) {
        const uint8_t dcnum = dclist[i];
        if(r->valid[dcnum]) {
            const uint8_t* dname = r->mappings[dcnum].cname;
            dname_copy(result->dname, dname);
            if(dname_status(dname) == DNAME_PARTIAL) {
                dname_cat(result->dname, origin);
                assert(dname_status(dname) == DNAME_VALID);
            }
            return;
        }
    }

    // We cannot reach this.  Config-time code ensures every resource
    //  has at least one valid datacenter+address, which will be picked
    //  up in the loop above and trigger a return
    assert(0);
}
